import 'package:flutter/material.dart';
import 'package:flutter/foundation.dart';
import 'iroute_master.dart';
import 'url_parser.dart';
import 'page_list.dart';

typedef RouteBuilder = Widget Function(
    BuildContext context, Map<String, dynamic> params);

/// The RouteDelegate defines application specific behaviors of how the router
/// learns about changes in the application state and how it responds to them.
/// It listens to the RouteInformation Parser and the app state and builds the Navigator with
/// the current list of pages (immutable object used to set navigator's history stack).
class IRoute extends RouterDelegate<String>
    with ChangeNotifier, PopNavigatorRouterDelegateMixin<String> {
  //This is the state of the navigator widget (in build method).
  //GlobalKey<NavigatorState> get navigatorKey => GlobalKey<NavigatorState>();

  GlobalKey<NavigatorState> get navigatorKey =>
      GlobalObjectKey<NavigatorState>(this);

  static IRoute of(BuildContext context) {
    final delegate = Router.of(context).routerDelegate;
    assert(delegate is IRoute, 'Delegate type must match');
    return delegate as IRoute;
  }

  late Map<String, RouteBuilder> routes;
  //just set the url
  //we know who is the url in routes
  final String defaultUrl; //默认的主页为/
  final String unknowUrl; //默认的未找到页面url为404
  IRouteMaster? _root;
  bool get isRoot => _root == null;
  set root(value) {
    assert(value != null);
    this._root = value;
    _root!.nested = this;
    _root!.addListener(this.notifyListeners);
  }

  PageList? _pageList;

  set pageList(PageList value) {
    this._pageList = value;
  }

  int get selectedPage {
    return _pageList!.current;
  }

  //切换页面，这里应处理页面build
  //这个属性仅供shell使用
  set selectedPage(int value) {
    _pageList!.current = value;
    push(_pageList!.pages[_pageList!.current].currentUrl);
    //接下来，默认url，当前url，当前页面组件
  }

  @override
  void dispose() {
    super.dispose();
    if (!isRoot) _root!.removeListener(this.notifyListeners);
  }

  IRoute(
      {required this.routes,
      this.defaultUrl = '/',
      this.unknowUrl = '/404',
      IRouteMaster? root})
      : _root = root ?? root {
    //
    if (routes.containsKey('_')) {
      addHome = (context, params) => routes['_']!(context, params);
    }
  } //指定初始的url？

  RouteBuilder? addHome;

  /// Internal backstack and pages representation.
  List<String> _routeStack = [];
  List<String> get routeStack => _routeStack;

  bool _canPop = true;
  bool get canPop {
    if (_canPop == false) return false;

    return _routeStack.length > 1;
  }

  set canPop(bool canPop) => _canPop = canPop;

  /// CurrentConfiguration detects changes in the route information
  /// It helps complete the browser history and enables browser back and forward buttons.
  String? get currentConfiguration {
    if (isRoot) {
      var result = routeStack.isNotEmpty ? routeStack.last : null;
      return result;
    }
  }

  //这是build函数中使用的，和push、切换页面无关
  //由于只有push时需要为url加入ipage参数，这里不用管

  //by url,get the WidgetBuilder and arguments
  Widget _getWidget(context, url) {
    String pattern = UrlParser.getPattern(url);
    var params = UrlParser.getParam(url);

    //print('pattern=$pattern,params length=${params.length}');
    // Widget result =
    //     ((context) => this.routes[pattern]!(context, params))(context);

    //在这里设置current，setNewPath的page就不会改变
    if (!isRoot) {
      //注意这里不能触发重绘，因此，不可使用selectedPage属性
      if (params.containsKey('page'))
        _pageList!.current = params['page'];
      else
        _pageList!.current = 0;

      //如果当前页面，其widget为null，或者当前页面的url不等于传入的url
      //就需要设置组件(
      if (_pageList!.pages[_pageList!.current].widget == null ||
          _pageList!.pages[_pageList!.current].currentUrl != url) {
        if (addHome != null)
          _pageList!.pages[_pageList!.current].widget = addHome!(context, {
            'child':
                ((context) => this.routes[pattern]!(context, params))(context)
          });
        else
          _pageList!.pages[_pageList!.current].widget =
              ((context) => this.routes[pattern]!(context, params))(context);
      }
      //在这里变动currentUrl
      if (_pageList!.pages[_pageList!.current].currentUrl != url)
        _pageList!.pages[_pageList!.current].currentUrl = url;

      //然后返回该组件
      return _pageList!.pages[_pageList!.current].widget!;
    } else
      return ((context) => this.routes[pattern]!(context, params))(context);
  }

  //根据完整url构建页面
  Page buildPage(context, url) {
    if (url == this.defaultUrl)
      return MaterialPage(
          key: UniqueKey(),
          name: '/',
          child: this._getWidget(
              context, this.defaultUrl) //add argument like(id:1) here
          );
    if (url == this.unknowUrl)
      return MaterialPage(
          key: UniqueKey(), name: url, child: this._getWidget(context, url));
    return MaterialPage(
        key: UniqueKey(), name: url, child: _getWidget(context, url));
  }

  @override
  Widget build(BuildContext context) {
    if (this._routeStack.isEmpty) _routeStack.add(this.defaultUrl);
    List<Page<dynamic>> _pages = [];
    //if (routeStack.isEmpty) _pages.add(buildPage(context, '/'));
    for (String path in routeStack) _pages.add(buildPage(context, path));

    return Navigator(
        key: navigatorKey,
        pages: _pages,
        onPopPage: (route, result) {
          print('Pop Page');
          if (!route.didPop(result)) {
            return false;
          }
          if (canPop) {
            pop();
          }
          return true;
        });
  }

  @override
  Future<void> setNewRoutePath(String path) async {
    if (_canPop == false) return;
    if (path == currentConfiguration) return;
    _routeStack = _setNewRouteHistory(_routeStack, path);
    if (!isRoot) {
      var params = UrlParser.getParam(path);
      if (params.containsKey('page'))
        _pageList!.current = params['page'];
      else
        _pageList!.current = 0;
      //这里需要修改页号？
      _root!.currentConfiguration = this._routeStack.last;
    } else
      notifyListeners();

    return;
  }

  @override
  Future<bool> popRoute() {
    print('Pop Route');
    return super.popRoute();
  }

  /// Updates route path history.
  ///
  /// In a browser, forward and backward navigation
  /// is indeterminate and a custom path history stack
  /// implementation is needed.
  /// When a [newRoute] is added, check the existing [routes]
  /// to see if the path already exists. If the path exists,
  /// remove all path entries on top of the path.
  /// Otherwise, add the new path to the path list.
  List<String> _setNewRouteHistory(List<String> routes, String newRoute) {
    List<String> pathsHolder = [];
    pathsHolder.addAll(routes);
    // Check if new path exists in history.
    for (String path in routes) {
      // If path exists, remove all paths on top.
      if (path == newRoute) {
        int index = routes.indexOf(path);
        int count = routes.length;
        for (var i = index; i < count - 1; i++) {
          pathsHolder.removeLast();
        }
        return pathsHolder;
      }
    }

    // Add new path to history.
    pathsHolder.add(newRoute);

    return pathsHolder;
  }

  void push(String path) {
    if (path == _routeStack.last) return;
    String url = path;

    if (_pageList != null) {
      //加上&& _pageList!.current != 0，则0号页面不需要加入ipage
      if (!path.contains(':ipage') && path != '/') {
        url = url + ':ipage=${_pageList!.current}';
      }
    }

    _routeStack.add(url);
    //notifyListeners();
    if (!isRoot)
      _root!.currentConfiguration = this._routeStack.last;
    else
      notifyListeners();
  }

  void pop() {
    _routeStack.removeLast();
    if (!isRoot) {
      var params = UrlParser.getParam(_routeStack.last);
      if (params.containsKey('page'))
        _pageList!.current = params['page'];
      else
        _pageList!.current = 0;
      _root!.currentConfiguration = this._routeStack.last;
    } else
      notifyListeners();
  }
}
